/*
 * OFAILocalNER.java
 * 
 * Copyright (c) 2006, 2007, Austrian Research Institute for Artificial Intelligence 
 *
 * This program is based on DefaultGazetteer.java of GATE which is 
 * Copyright (c) 1998-2005, The University of Sheffield.
 *
 * This is free software, licenced under the GNU Library General Public License,
 * Version 2, June1991. See http://www.gnu.org/licenses/gpl.txt
 * 
 * Johann Petrak, 2006-03, updated 2007-03
 *
 * in large parts based on code by 
 * Valentin Tablan, 03/07/2000
 * borislav popov 24/03/2002
 *
 */
package at.ofai.gate.localner;

import java.util.*;
import java.util.regex.*;
import java.net.URL;

import gate.*;
import gate.creole.*;
//import gate.event.*;
//import gate.creole.gazetteer.*;
//import gate.util.*;
import java.io.*;

import org.openrdf.repository.Repository;
import org.openrdf.repository.http.HTTPRepository;
import org.openrdf.repository.RepositoryException;
import org.openrdf.OpenRDFException;
import org.openrdf.repository.RepositoryConnection;
import org.openrdf.query.TupleQuery;
import org.openrdf.query.TupleQueryResult;
import org.openrdf.query.BindingSet;
import org.openrdf.query.QueryLanguage;
import org.openrdf.model.Value;
/** This is ..
 * explain!
 *  @author Johann Petrak
 */
public class OFAILocalNER
    extends gate.creole.AbstractLanguageAnalyser
    implements ControllerAwarePR
{

  private static final long serialVersionUID = 1122934431234571432L;

  
  static boolean isShownVersion = false;
  
  // compile a pattern for strings starting or ending in punctuations!
  private static final String PUNCTSTART = "^\\p{Punct}";
  private static final String PUNCTEND   = "\\p{Punct}$";
  private static Pattern punctStartPattern = Pattern.compile(PUNCTSTART);
  private static Pattern punctEndPattern   = Pattern.compile(PUNCTEND);

  
  protected URL   queryFileURL;
  public void setQueryFileURL(URL queryFileURL) {
    this.queryFileURL = queryFileURL;
  }
  public URL getQueryFileURL() {
    return queryFileURL;
  }

  protected String queryString;
  public void setQueryString(String queryString) {
    this.queryString= queryString;
  }
  public String getQueryString() {
    return queryString;
  }

  protected URL  sesameServerURL;
  public void setSesameServerURL(URL sesameServerURL) {
    this.sesameServerURL = sesameServerURL;
  }
  public URL getSesameServerURL() {
    return sesameServerURL;
  }

  protected String sesameRepositoryID;
  public void setSesameRepositoryID(String sesameRepositoryID) {
    this.sesameRepositoryID = sesameRepositoryID;
  }
  public String getSesameRepositoryID() {
    return sesameRepositoryID;
  }

  protected String annotationSetName;
  public void setAnnotationSetName(String annotationSetName) {
    this.annotationSetName = annotationSetName;
  }
  public String getAnnotationSetName() {
    return annotationSetName;
  }

  // the annotation type to assign to local NERs 
  protected String annotationType;
  public void setAnnotationType(String annotationType) {
    this.annotationType = annotationType;
  }
  public String getAnnotationType() {
    return annotationType;
  }

  protected String docIdAnnSet;
  public void setDocIdAnnSet(String docIdAnnSet) {
    this.docIdAnnSet = docIdAnnSet;
  }
  public String getDocIdAnnSet() {
    return docIdAnnSet;
  }

  protected String docIdAnnType;
  public void setDocIdAnnType(String docIdAnnType) {
    this.docIdAnnType = docIdAnnType;
  }
  public String getDocIdAnnType() {
    return docIdAnnType;
  }

  protected String docIdFeature;
  public void setDocIdFeature(String docIdFeature) {
    this.docIdFeature = docIdFeature;
  }
  public String getDocIdFeature() {
    return docIdFeature;
  }

  protected String baseURI;
  public void setBaseURI(String baseURI) {
    this.baseURI = baseURI;
  }
  public String getBaseURI() {
    return baseURI;
  }

  protected Boolean wholeWordsOnly;
  public void setWholeWordsOnly(Boolean yesorno) {
      wholeWordsOnly = yesorno;
  }
  public Boolean getWholeWordsOnly() {
      return wholeWordsOnly;
  }

  protected Boolean ignoreCase;
  public void setIgnoreCase(Boolean yesorno) {
      ignoreCase = yesorno;
  }
  public Boolean getIgnoreCase() {
      return ignoreCase;
  }

  protected Integer maxResults;
  public void setMaxResults(Integer max) {
      maxResults = max;
  }
  public Integer getMaxResults() {
      return maxResults;
  }

  public void controllerExecutionStarted(gate.Controller c)
    throws ExecutionException {
    startup();
  }
  public void controllerExecutionFinished(gate.Controller c)
    throws ExecutionException {
    shutdown();
  }
  public void controllerExecutionAborted(gate.Controller c, java.lang.Throwable t)
    throws ExecutionException {
    
  }

  private String mQuery;

  Repository sesameRepository;
  /** 
   * Build a gazetter using the default lists from the gate resources
   */
  public OFAILocalNER(){
  }

  public void startup() throws ExecutionException {
    // take query from input or read it from file 
    System.out.println("This is the startup for the corpus");

    if (! queryString.equals("")) {
	mQuery = queryString;
    } 
    else {
    String tmp = "";
    try {
      tmp = queryFileURL.getProtocol();
      if(!tmp.equals("file")) {
        throw new ExecutionException("Invalid protocol for query File URL: "+tmp);
      }
      tmp = queryFileURL.getPath();
      mQuery = readTextFile(tmp,"utf-8");
    } catch (IOException e) {
      throw new ExecutionException("Could not open file " + tmp + ": " + e);
    }
    }
    System.out.println("Trying to connect to Sesame repository: "+sesameServerURL+"/"+sesameRepositoryID);
    try {
      sesameRepository = new HTTPRepository(sesameServerURL.toString(), sesameRepositoryID);
      sesameRepository.initialize();
    } catch (RepositoryException e) {
      System.err.println("Exception when initializing repository: " + e);
    }
 }
  
  public void shutdown() {
    // TODO: disconnect Sesame
    System.out.println("This is the shutdown method of OFAILocalNER");
  }

  @Override
  public Resource init() throws ResourceInstantiationException {
    /* do the init stuff we need specially for this subclass */
    if(!isShownVersion) {
      System.out.println("Plugin version: $Id: OFAILocalNER.java 439 2009-03-25 20:18:51Z hannes $");
      isShownVersion = true;
    }
    return super.init();
  }
  

  /**
   * This method runs the gazetteer.
   */
    
  //@throws gate.creole.ExecutionException;
  public void execute() throws ExecutionException{
    interrupted = false;
    AnnotationSet annotationSet;  // where to place the generated anns
    AnnotationSet docIdSet;       // where to find the document id
    String thisQuery = mQuery;    // the query modified for this document
    //check the input
    if(document == null) {
      throw new ExecutionException(
        "No document to process!"
      );
    }

    // First get the id feature .. if we do not find it, we can skip
    // processing this document
    // 1-get the set
    if(docIdAnnSet == null || docIdAnnSet.equals("")) {
      docIdSet = document.getAnnotations();
    } else {
      docIdSet = document.getAnnotations(docIdAnnSet);
    }
    // filter the type
    // we assume that there is only one annotation of this max, if there
    // are more, we take whatever the iterator returns first
    // If there is none, we are done.
    docIdSet = docIdSet.get(docIdAnnType);
    if(docIdSet.size() == 0) {
      System.out.println("No ID found for document");
      return;
    }
    Annotation ann = docIdSet.iterator().next();
    String id = (String)ann.getFeatures().get(docIdFeature);

    if(id == null) {
      System.out.println("No ID found for document");
      return;
    }
    // modify the query by replacing the string "%id%" with the actual
    // document id
    System.out.println("Document id="+id);

    thisQuery = thisQuery.replaceAll("%id%", id);
    System.out.println("Using query:"); System.out.println();
    System.out.println(thisQuery);
    // query the repository -- if we do not find any results, we are done
    // we fetch the following things from the result:
    // - the name of the second binding which will be used as the feature name
    // - a map that maps mention strings to sets of class strings
    // change 27/02/2009: do not insist on exactly 2 bindings but on any number >= 1 
    // i.e. allow from 0 to n features to be set. 
    //hp
    HashMap<String,Vector<String[]>>  mentionMap =
            new HashMap<String,Vector<String[]>>();
    
    // put binding names to global variable
    String[] bNames;  // bindingNames.toArray(new String[0]);

    int n = 0;
    try {
      //System.err.println("Have the repository connection");
      RepositoryConnection con = sesameRepository.getConnection();
      TupleQuery tupleQuery = con.prepareTupleQuery(QueryLanguage.SPARQL, thisQuery);
      TupleQueryResult result = tupleQuery.evaluate();
      System.out.println("Have the query evaluated: ");
      try {
        List<String> bindingNames = result.getBindingNames();
        bNames = bindingNames.toArray(new String[0]);

        // ## For now, only accept exactly two bindings here
        // from now on allow any number of bindings >= 2
        if(bNames.length == 0) {
          throw new ExecutionException("Query must return at least one variable");
        }
        // the second (and all following)  binding names is what we will use for the feature name
        //System.err.println("Have a result");
        while (result.hasNext() && n < maxResults) {
          //System.err.println("Have a result row");
          BindingSet bindingSet = result.next();
          Value value = bindingSet.getValue(bindingNames.get(0));
          String mention = value.stringValue();
          if(ignoreCase) {
            mention = mention.toLowerCase();
          }
	  // iterate over bindings 
	  // array for storing feature-values: the array is bNames-1 long because bNames[0] is NOT used
	  String[] featureValues = new String[bNames.length - 1];
	  // bindingNames from 1 to n will contain feature-names 
	  for (int i = 1; i < bNames.length; i++) {
	       value = bindingSet.getValue(bindingNames.get(i));
	      // hp // featureValues: was classname 
	       // it is possible that sparql returns some undefined values:
	       if (value == null) {
		   featureValues[i-1] = "$$FEATURENOTFOUND$$";
		   System.err.println("Warning: SPARQL did not provide value for feature " + bNames[i]);
	       }
	       else {
		   featureValues[i-1] = value.stringValue();
	       }
	  }    


          // check if that mention is already in the map: if yes,
          // add the featureValues  to the Vector, otherwise create a new Vector
          Vector<String[]> hs = mentionMap.get(mention); //hp
          if(hs == null) {
            hs = new Vector<String[]>();
          }
          hs.add(featureValues);
          mentionMap.put(mention,hs);
          n++;
        }  // while has result
        System.out.println("Found results: "+n);
      } finally {
        result.close();
        con.close();
      }
    } catch(OpenRDFException e) {
      throw new ExecutionException("Error opening RDF: "+e);
    }
    if(n==0) {
      System.err.println("No result for this doc, doing nothing");
      return;
    }
    // if we have a result, get the annotation set where to put the
    // lookup annotations in...
    if(annotationSetName == null ||
       annotationSetName.equals("")) {
      annotationSet = document.getAnnotations();
    } else {
      annotationSet = document.getAnnotations(annotationSetName);
    }

    fireStatusChanged("Performing look-up in " + document.getName() + "...");
    String content = document.getContent().toString();
    // int length = content.length();

    // create a single pattern from all the strings we got and find all
    // occurrences.
    StringBuffer patternstring = new StringBuffer();
    // Iterator<String> i = mentionMap.keySet().iterator();
    boolean first = true;
    Matcher punctStartMatcher = null;
    Matcher punctEndMatcher   = null;
    for(String s : mentionMap.keySet()) {
      // TEST: is there a punctuation at the start or the end of s?
      punctStartMatcher = punctStartPattern.matcher(s);
      punctEndMatcher   = punctEndPattern.matcher(s);
      if(first) {
        first = false;
      } else {
        patternstring.append("|");
      }
      s="\\Q"+s+"\\E"; // quote the string
      
      if(wholeWordsOnly) {
          // add \\b 
          // BUT:
          // if s starts or ends  in a punctuation (e.g. '#1' or 'Against Me!') 
          // then \\b will prevent any match!
          // fix:  in this case omit \\b 
          if (punctStartMatcher.find()) {
              System.out.println("INITIAL_PUNCTUATION found: "+s);
          }  else {
              s="\\b"+s; 
                  }
          if (punctEndMatcher.find()) {
              System.out.println("FINAL_PUNCTUATION found: "+s);
          }  else { 
              s=s+"\\b"; 
          }
      }
      patternstring.append(s);
    }
    System.out.println("Have created the pattern: "+patternstring.toString());
    int compileFlags = java.util.regex.Pattern.CANON_EQ;
    if(ignoreCase) {
        compileFlags = compileFlags | java.util.regex.Pattern.CASE_INSENSITIVE;
    }
    Pattern p = Pattern.compile(patternstring.toString(),compileFlags);

    Matcher m = p.matcher(content);
    while(m.find()) {
      int from = m.start();
      int to   = m.end();
      String match = content.substring(from,to);
      String matchlower = match.toLowerCase();
      System.out.println("Found a match: "+match+", lower case is: "+matchlower);
      if(ignoreCase) {
        match = matchlower;
      }
      // NOTE: this uses the generic "toLowerCase" on both the search string
      // and the found string to store and later look up in the hash map
      // However, the way case insensitive, unicode-aware matching is done
      // in the regexp will differ and the regexp might find matches that
      // are different than just the lowercase version of the search string.
      // This will result in the "Odd error" below -- the solution should be
      // to 1) in such a case do linear search using regexp matching to
      // look up the class and 2) assign some default class if we still fail.

      // HashSet<String> classnames = mentionMap.get(match);
      Vector<String[]> featureLists = mentionMap.get(match); //hp
      if(featureLists == null) {
        System.err.println("Odd error - cannot find match "+match+" in hashmap!");
        featureLists = new Vector<String[]>(); 
        String[] errorFeatList = new String[bNames.length - 1];
        for (int k = 0; k <  errorFeatList.length; k++) {
            errorFeatList[k] = "$$FEATURENOTFOUND$$";
        }
            featureLists.add(errorFeatList);
      }
      
      // featureLists is a list of arrays of strings
      ListIterator<String[]> fli = featureLists.listIterator();
      while (fli.hasNext()) {
	  String[] featList = (String[]) fli.next();
          // for each collection of features we found, create an annotation
	  FeatureMap fm = Factory.newFeatureMap();
	  for (int j = 0; j < featList.length; j++) {
	      String featureName  =  bNames[j+1];
	      String featureValue =  featList[j];
	      fm.put(featureName, featureValue);
	  }
	  try {
	      annotationSet.add(new Long(from),
				new Long(to),
				annotationType,
				fm);
	  } catch (gate.util.InvalidOffsetException e) {
	      System.err.println("Invalid offset exception when annotating "+match+" in doc "+document.getName());
	  }
      }
    }
    // assign annotations for each unique class we got from the result
    // set

    fireProcessFinished();
    fireStatusChanged("Lookup complete!");
} // execute

  /** A set containing all the states of the FSM backing the gazetteer
   */
 // Set fsmStates; // not used?
private static String readTextFile(String f, String encoding)
          throws IOException {
    Reader r = new InputStreamReader(
            new FileInputStream(f), encoding);
    try {
      StringWriter w = new StringWriter();
      char[] buffer = new char[65536];

      int len;
      while ((len = r.read(buffer)) != -1) {
        w.write(buffer, 0, len);
      }

      return w.toString();
    } finally {
      r.close();
    }
  }
} // DefaultGazetteer

